import logging
import os
import sys


def isfile(path):
    """
    Check if file exists.

    Args:
        path (_str_): Path to file

    Returns:
        _str_: Returns abspath to file if found, else prints warning and exits.
    """
    if not os.path.isfile(path):
        logging.error('Input file not found: %s' % path)
        sys.exit(1)
    else:
        return os.path.abspath(path)



from itertools import groupby

from teloclip.motifs import check_sequence_for_patterns
from teloclip.utils import isfile


def makeMask(killIdx, listlen):
    # makeMask([0,9], 10) =  [0,1,1,1,1,1,1,1,1,0]
    mask = [1 for i in range(listlen)]
    for x in killIdx:
        mask[x] = 0
    return mask


# NCU
def filterList(data, exclude):
    # filterList([1,2,3,4,5,6,7,8,9,10],[0,9]) = [2,3,4,5,6,7,8,9]
    mask = makeMask(exclude, len(data))
    return (d for d, s in zip(data, mask) if s)


# NCU
def revComp(seq):
    """Rev comp DNA string."""

    def revcompl(x):
        return ''.join(
            [{'A': 'T', 'C': 'G', 'G': 'C', 'T': 'A', 'N': 'N'}[B] for B in x][::-1]
        )

    return revcompl(seq)


# NCU
def writeClip(idx, zpad, gap, seq, maplen):
    # leftpad idx ID
    padIdx = str(idx).zfill(zpad) + ':'
    # If gap between aln end(R) or start(L) and contig end, left pad softclip with '-'
    padseq = '-' * gap + seq
    # Format length of ref covered by alingment
    readlen = 'LEN=' + str(maplen).rjust(6)
    print('\t'.join([padIdx, readlen, padseq]))


# NCU
def fasta2dict(fasta_name):
    fh = open(fasta_name)
    faiter = (x[1] for x in groupby(fh, lambda line: line[0] == '>'))
    contigDict = {}
    for header in faiter:
        # Drop the ">"
        # Split on whitespace and take first item as name
        header = header.__next__()[1:].strip()
        name = header.split()[0]
        # Join all sequence lines to one.
        seq = ''.join(s.strip() for s in faiter.__next__())
        contigDict[name] = (header, seq)
    return contigDict


def writefasta(outfile, name, seq, length=80):
    outfile.write('>' + str(name) + '\n')
    while len(seq) > 0:
        outfile.write(seq[:length] + '\n')
        seq = seq[length:]


def read_fai(fai):
    """
    Import fasta index file. Return dict of sequence names and lengths.
    """
    path = isfile(fai)
    # Init empty dict
    ContigDict = {}
    # Read fai_file to dict
    with open(path, 'r') as f:
        for line in f.readlines():
            li = line.strip().split()
            ContigDict[li[0]] = int(li[1])
    return ContigDict


def addRevComplement(motifList):
    """
    Take list of DNA motif strings and return unique set of strings and their reverse complements.
    """

    def revcompl(x):
        return ''.join(
            [{'A': 'T', 'C': 'G', 'G': 'C', 'T': 'A', 'N': 'N'}[B] for B in x][::-1]
        )

    setList = []
    for motif in motifList:
        setList.append(motif)
        setList.append(revcompl(motif))
    return set(setList)


def isMotifInClip(
    samline, motifList, leftClip, rightClip, leftClipLen, rightClipLen, minRepeats=1
):
    """
    Extract terminal soft-clipped blocks from read sequence and test for presence of any DNA motif in motifList.
    """
    # Sam seq field
    SAM_SEQ = 9

    # Initialize leftcheck and rightcheck
    leftcheck = False
    rightcheck = False

    # Search motif/s as regex in the clipped segment
    if leftClip:
        leftcheck = check_sequence_for_patterns(
            samline[SAM_SEQ][0:leftClipLen], motifList, minRepeats
        )
    if rightClip:
        rightcheck = check_sequence_for_patterns(
            samline[SAM_SEQ][-rightClipLen:], motifList, minRepeats
        )

    # True if either clipped end sequence contains at least one instance of any motif.
    return any([leftcheck, rightcheck])



"""
                                       ██████╗ ██╗      ██╗ ██████╗
                                      ██╔════╝ ██║      ██║ ██╔══██╗
                                      ██║      ██║      ██║ ██████╔╝
████████╗ ███████╗ ██╗       ██████╗  ██║      ██║      ██║ ██╔═══╝
╚══██╔══╝ ██╔════╝ ██║      ██╔═══██╗ ╚██████╗ ███████╗ ██║ ██║
   ██║    █████╗   ██║      ██║   ██║  ╚═════╝ ╚══════╝ ╚═╝ ╚═╝
   ██║    ██╔══╝   ██║      ██║   ██║
   ██║    ███████╗ ███████╗ ╚██████╔╝
   ╚═╝    ╚══════╝ ╚══════╝  ╚═════╝

A tool for the recovery of unassembled telomeres from soft-clipped read alignments.

"""

import argparse
import logging
import sys

from teloclip._version import __version__
from teloclip.logs import init_logging
from teloclip.motifs import make_fuzzy_motif_regex, make_motif_regex
from teloclip.samops import processSamlines
from teloclip.seqops import addRevComplement, read_fai


def mainArgs():
    parser = argparse.ArgumentParser(
        description='Filter SAM file for clipped alignments containing unassembled telomeric repeats.',
        prog='teloclip',
    )
    # Input options
    parser.add_argument(
        'samfile', nargs='?', type=argparse.FileType('r'), default=sys.stdin
    )
    parser.add_argument(
        '--ref-idx',
        dest='ref_idx',
        type=str,
        required=True,
        help='Path to fai index for reference fasta. Index fasta using `samtools faidx FASTA`',
    )
    parser.add_argument(
        '--min-clip',
        dest='min_clip',
        type=int,
        default=1,
        help='Require clip to extend past ref contig end by at least N bases.',
    )
    parser.add_argument(
        '--max-break',
        dest='max_break',
        type=int,
        default=50,
        help='Tolerate max N unaligned bases before contig end.',
    )
    parser.add_argument(
        '--motifs',
        type=str,
        default=None,
        help='If set keep only reads containing given motif/s from comma delimited list of strings. '
        'By default also search for reverse complement of motifs. i.e. TTAGGG,TTAAGGG will also match CCCTAA,CCCTTAA',
    )
    parser.add_argument(
        '--no-rev',
        dest='no_rev',
        default=False,
        action='store_true',
        help='If set do NOT search for reverse complement of specified motifs.',
    )
    parser.add_argument(
        '--fuzzy',
        default=False,
        action='store_true',
        help='If set, tolerate +/- 1 variation in motif homopolymer runs '
        'i.e. TTAGGG -> T{1,3}AG{2,4}. Default: Off',
    )
    parser.add_argument(
        '-r',
        '--min-repeats',
        dest='min_repeats',
        default=1,
        type=int,
        help='Minimum number of sequential pattern matches required for a hit to be reported. Default: 3',
    )
    parser.add_argument(
        '--min-anchor',
        dest='min_anchor',
        default=500,
        type=int,
        help='Minimum number of aligned bases (anchor) required on the non-clipped portion of the read. Default: 500',
    )
    parser.add_argument(
        '--match-anywhere',
        dest='match_anywhere',
        default=False,
        action='store_true',
        help='If set, motif match may occur in unclipped region of reads.',
    )
    parser.add_argument(
        '-v',
        '--version',
        action='version',
        version='%(prog)s {version}'.format(version=__version__),
    )
    args = parser.parse_args()

    return args


def main():
    # Set up logging
    init_logging()

    # Get cmd line args
    args = mainArgs()

    # Fetch contig lens
    contig_dict = read_fai(args.ref_idx)

    if args.motifs:
        # Create list of motifs from comma delimited string
        motifList = args.motifs.split(',')

        # Add rev comp motifs to list, make unique set.
        if not args.no_rev:
            motifList = addRevComplement(motifList)

        # Make unique set.
        motifList = list(set(motifList))

        # Create regex patterns for each motif
        if args.fuzzy:
            # Create fuzzy regex patterns for each motif
            logging.info(
                'Using option --fuzzy to find inexact motif matches. Tolerate +/- 1 base variance in motif homopolymers.'
            )
            motifList = [make_fuzzy_motif_regex(motif) for motif in motifList]
        else:
            # Create exact regex patterns for each motif
            motifList = [make_motif_regex(motif) for motif in motifList]
    else:
        motifList = []

    # Process SAM lines

    processSamlines(
        args.samfile,
        contig_dict,
        motifList,
        match_anywhere=args.match_anywhere,
        max_break=args.max_break,
        min_clip=args.min_clip,
        min_repeats=args.min_repeats,
        min_anchor=args.min_anchor,
    )
